# Ark programming language

Ark is a toy programming language, inspired from Lisp. The language is dynamically and strongly typed, this means that a variable can not be implicitely converted to another type, and that any variable can hold any type.

The language itself is very small since it has only 8 keywords: `begin`, `if`, `while`, `let`, `mut`, `set`, `fun`, `import`, `quote` and `del`.

## Basics

Blocks are surrounded by `()`. Creating a block with `[...]` will expand to `(list ...)`, and creating one with `{...}` to `(begin ...)`.

`1257756` and `-1.2` are numbers.

`"hello world"` is a string.

A comment starts with `#`.

Instead of calling a function as you would do in C or Python, in Ark we use the Polish notation, so `f(a, b, c, ...)` becomes `(f a b c ...)`.

### Hello world!

```
{
    (let message "hello world!")
    (let sayHi! (fun (msg)
        (print msg)
    ))
    (sayHi! message)  # display `hello world!`
}
```

## Creating a variable

A variable is available in the lexical scope it was created in. In C-like syntax:

```c
void fun()
{
    int x = 5;

    void fun2()
    {
        printf("%d", x);
    }
}
```

Every inner level can access its outer levels.

To define a new immutable variable, we use the `let` keyword, taking a symbol and a value:

```clojure
(let a 10)  # valid definition
(let d 1 2)  # invalid definition
(let c) # invalid definition
```

To be able to modify a variable, we must declare it as **mutable** using the `mut` keyword. Then we can change its value using `set`:

```
(mut a 5)
(set a 12)  # valid
(set b 1 2)  # invalid
(set c)  # invalid
```

## Conditions

A condition can be created using `if` keyword, taking 3 nodes: a condition, a if block, and a else block.

```clojure
(if (= a 10)
    (print "a = 10")  # then
    (print "a != 10") # else
)
```

## Blocks

Sometimes we need to have a single block doing multiple things. We could create a function but it's too heavy, so we create blocks:

```clojure
(if (= a 10)
    # then
    (begin
        (set a 12)
        (print "because a = 10, I changed it to 12")
    )

    # else
    (print "nope")
)
```

## Loops

They can be summoned with `while`, taking two nodes: a condition and a "then" block.

```clojure
(while (< a 10)
    (begin
        (set a (+ a 1))
        (print a)
    )
)
```

## Functions and closures

We can create a function with the keyword `fun`. It needs 2 nodes: an argument list, and the body.

```clojure
(fun (a b c) (print a b c))
```

When creating a function, it's capturing the environment it was created in, and copying it to save it. We create what we call a "closure".

```clojure
(let make-closure (fun (name)
    (fun ()
        (print name)
    )
))

# instanciating a closure
(let say-hi (make-closure "hi"))
# calling our closure
(say-hi)  # prints `hi`
```

**Nota Bene**: all the arguments passed to a function are passed by value, and they are marked as *mutable* inside the function.

## Importing Ark code

To import an Ark file into your code you can do this: `(import "myfile.ark")`. The function needs a path relative to the location of the code importing the Ark files, and takes only one argument.

If you want to import an Ark file from the standard library, `import` will be able to find it, no need to create a relative path to the lib! Just note that all the files in the standard library have a name starting with an uppercase letter, for example `Switch.ark`.

## Loading a plugin

You may want to be able to use the SFML in Ark, or another C++ library. If an Ark plugin exists to use this library (or if you made your own), you just have to put the module alongside the `.arkc` file generated by the VM, and use the keyword `import` in your code, as follows: `(import "mymodule.so")`. `import` will automatically detect if you are loading a plugin, based on the extension of the file.

The files are searched in the directory when the `.arkc` file is, and in the directory where the Ark standard library is installed.

## Defering a function call

```clojure
{
    (let foo (fun (n) (print n)))
    (let egg (quote (foo 5)))
    (print egg)  # prints `Closure @ ...`
    (egg)  # prints `5`
}
```

The `'(foo 5)` notation can also be used, it will be automatically expanded to `(quote ...)`.

# Using Ark in a project

-> [Embedding Ark](embedding.md)

# Standard library

-> [Ark std lib](lib.md)

# The bytecode

-> [Ark bytecode format](bytecode.md)
